#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from __future__ import annotations
from argparse import ArgumentParser
import logging
from typing import Optional


parser = ArgumentParser()
parser.add_argument("-r", "--host", help="set the host name of this Node's Coordinator")
parser.add_argument("-n", "--name", help="set the application name")
parser.add_argument("-q", "--quiet", action="count", default=0,
                    help="decrease the logging level by one, may be used more than once")
parser.add_argument("-v", "--verbose", action="count", default=0,
                    help="increase the logging level by one, may be used more than once")


def parse_command_line_parameters(parser: ArgumentParser = parser,
                                  logger: Optional[logging.Logger] = None,
                                  arguments: Optional[list[str]] = None,
                                  parser_description: Optional[str] = None,
                                  logging_default: int = logging.WARNING,
                                  ) -> dict:
    """Parse the command line parameters and return a dictionary for GUIs.

    :param parser: parser to use, for example with more settings.
    :param logger: The logger whose log level to set. Defaults to "__main__" logger.
    :param list arguments: Arguments for the parser to parse. Per default, take it from `sys.argv`.
    :param str parser_description: Override the parsers program description description.
    :param int logging_default: Default level for logging.
    :return: Dictionary with keyword arguments parsed from the command line parameters.
    """
    if parser_description is not None:
        parser.description = parser_description
    kwargs = vars(parser.parse_args(arguments))
    verbosity = logging_default + (kwargs.pop("quiet", 0) - kwargs.pop("verbose", 0)) * 10
    if logger is None:
        logger = logging.getLogger("__main__")
    logger.setLevel(verbosity)
    for key, value in list(kwargs.items()):
        # remove not set values
        if value is None:
            del kwargs[key]
    return kwargs



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from enum import IntEnum
import logging

from ..core.leco_protocols import LogLevels


class PythonLogLevels(IntEnum):
    """Attribution of Python log levels to the LECO ones."""
    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARNING = logging.WARNING
    ERROR = logging.ERROR
    CRITICAL = logging.CRITICAL


def get_leco_log_level(log_level: int) -> LogLevels:
    """Get the LECO log level from an integer (or python log level) if possible."""
    name = PythonLogLevels(log_level).name
    return LogLevels[name]



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

"""Hangs infinitely on import on github CI."""

# from __future__ import annotations
# import threading
# from typing import Optional

# import pytest
# import zmq

# from pyleco.core import (
#     PROXY_RECEIVING_PORT,
#     PROXY_SENDING_PORT,
# )
# from pyleco.test import FakeContext, FakeSocket

# from pyleco.coordinators.proxy_server import pub_sub_proxy, start_proxy

# parameters: tuple[FakeSocket, FakeSocket, Optional[FakeSocket]]


# @pytest.fixture
# def fake_proxy_steerable(monkeypatch: pytest.MonkeyPatch) -> None:
#     def _fake_proxy_steerable(
#         frontend: FakeSocket, backend: FakeSocket, capture: Optional[FakeSocket] = None
#     ):
#         global parameters
#         parameters = frontend, backend, capture
#         raise zmq.ContextTerminated

#     monkeypatch.setattr("zmq.proxy_steerable", _fake_proxy_steerable)


# class Test_pub_sub_proxy:
#     def test_default_config(self, fake_proxy_steerable):
#         pub_sub_proxy(FakeContext())  # type: ignore
#         global parameters
#         f, b, c = parameters
#         assert f.addr == f"tcp://*:{PROXY_SENDING_PORT}"
#         assert b.addr == f"tcp://*:{PROXY_RECEIVING_PORT}"
#         assert c is None

#     def test_event_set_for_successful_binding(self, fake_proxy_steerable):
#         event = threading.Event()
#         pub_sub_proxy(FakeContext(), event=event)  # type: ignore
#         assert event.is_set()

#     @pytest.mark.parametrize(
#         "pub, sub",
#         (
#             ("localhost", "remote"),
#             ("remote", "localhost"),
#             ("a", "b"),
#         ),
#     )
#     def test_remote_configuration(self, pub: str, sub: str, fake_proxy_steerable):
#         pub_sub_proxy(FakeContext(), sub=sub, pub=pub)  # type: ignore
#         global parameters
#         f, b, c = parameters
#         assert f.addr == f"tcp://{pub}:{PROXY_RECEIVING_PORT}"
#         assert b.addr == f"tcp://{sub}:{PROXY_SENDING_PORT}"

#     def test_capture(self, fake_proxy_steerable):
#         pub_sub_proxy(context=FakeContext(), captured=True)  # type: ignore
#         global parameters
#         f, b, c = parameters  # type: ignore
#         c: FakeSocket
#         assert c.addr == "inproc://capture"
#         assert c.socket_type == zmq.PUB


# def test_start_proxy():
#     context = start_proxy()
#     # assert no error is raised
#     context.term()


# @pytest.mark.skip(reason="Hangs infinitely")
# def test_start_proxy_fails_if_already_started():
#     # arrange
#     context = start_proxy()
#     with pytest.raises(TimeoutError):
#         start_proxy()
#     # assert no error is raised
#     context.term()



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from unittest.mock import MagicMock

import pytest
import zmq

from pyleco.core.message import Message
from pyleco.test import FakeContext

from pyleco.utils.pipe_handler import LockedMessageBuffer, PipeHandler, CommunicatorPipe,\
    PipeCommands

cid = b"conversation_id;"  # conversation_id
header = b"".join((cid, b"mid", b"\x00"))
# the result
msg = Message(b"r", b"s", conversation_id=cid, message_id=b"mid")
msg_list = ("r", "s", cid, b"", None)
# some different message
other = Message(b"r", b"s", conversation_id=b"conversation_id9", message_id=b"mid")


@pytest.fixture
def message_buffer() -> LockedMessageBuffer:
    message_buffer = LockedMessageBuffer()
    message_buffer._messages = [msg]
    return message_buffer


# Test MessageBuffer
def test_add_conversation_id(message_buffer: LockedMessageBuffer):
    message_buffer.add_conversation_id(conversation_id=cid)
    assert cid in message_buffer._requested_ids


class Test_add_response_message_successful:
    @pytest.fixture
    def message_buffer_added(self) -> LockedMessageBuffer:
        # Arrange
        mb = LockedMessageBuffer()
        assert len(mb) == 0
        mb.add_conversation_id(cid)
        # Act
        mb.add_message(msg)
        return mb

    def test_return_value(self, message_buffer_added: LockedMessageBuffer):
        assert len(message_buffer_added) == 1

    def test_msg_in_buffer(self, message_buffer_added: LockedMessageBuffer):
        assert message_buffer_added._messages == [msg]


class Test_check_message_in_buffer:
    @pytest.fixture
    def message_buffer_cmib(self, message_buffer: LockedMessageBuffer):
        predicate = message_buffer._predicate_generator(cid)
        message_buffer._predicate = predicate  # type: ignore
        return message_buffer

    def test_message_is_in_first_place(self, message_buffer_cmib: LockedMessageBuffer):
        assert message_buffer_cmib._predicate() == msg  # type: ignore
        assert message_buffer_cmib._messages == []

    def test_no_suitable_message_in_buffer(self, message_buffer: LockedMessageBuffer):
        predicate = message_buffer._predicate_generator(conversation_id=b"other_cid")
        assert predicate() is None
        assert message_buffer._messages != []

    def test_msg_somewhere_in_buffer(self, message_buffer_cmib: LockedMessageBuffer):
        o2 = Message(b"r", b"s", conversation_id=b"conversation_id9", message_id=b"mi7")
        message_buffer_cmib._messages = [other, msg, o2]
        assert message_buffer_cmib._predicate() == msg  # type:ignore
        assert message_buffer_cmib._messages == [other, o2]


@pytest.mark.parametrize("buffer", (
        [msg],  # msg is only message
        [msg, other],  # msg is in the first place of the buffer
        [other, msg],  # msg is in the second and last place of the buffer
        [other, msg, other]  # msg is in the middle of the buffer
    ))
def test_retrieve_message_success(message_buffer: LockedMessageBuffer, buffer):
    message_buffer._messages = buffer
    original_length = len(buffer)
    assert message_buffer.wait_for_message(cid) == msg
    assert len(message_buffer._messages) == original_length - 1


@pytest.mark.parametrize("buffer", (
        [],  # no message in buffer
        [other],  # other message in buffer
    ))
def test_retrieve_message_fail(message_buffer: LockedMessageBuffer, buffer):
    message_buffer._messages = buffer
    with pytest.raises(TimeoutError):
        message_buffer.wait_for_message(conversation_id=cid, timeout=0.01)


@pytest.mark.parametrize("length", (1, 3, 7))
def test_length_of_buffer(message_buffer: LockedMessageBuffer, length: int):
    message_buffer._messages = length * [msg]
    assert len(message_buffer) == length


# Test CommunicatorPipe
class Test_CommunicatorPipe_send_pipe:
    def test_send_pipe_message(self, communicator: CommunicatorPipe):
        communicator.socket.send_multipart = MagicMock()  # type: ignore[method-assign]
        communicator._send_pipe_message(PipeCommands.LOCAL_COMMAND, b"abc")
        # assert
        communicator.socket.send_multipart.assert_called_once_with(
            (PipeCommands.LOCAL_COMMAND, b"abc")
        )

    def test_raise_ConnectionError_on_zmq_error(self, communicator: CommunicatorPipe):
        communicator.socket.send_multipart = MagicMock(  # type: ignore[method-assign]
            side_effect=zmq.ZMQError(128, "not a socket")
        )
        # act
        with pytest.raises(ConnectionRefusedError):
            communicator._send_pipe_message(PipeCommands.LOCAL_COMMAND, b"c")


# Test PipeHandler
@pytest.fixture
def pipe_handler() -> PipeHandler:
    """With fake contexts, that is with a broken pipe."""
    pipe_handler = PipeHandler(name="handler", context=FakeContext())  # type: ignore
    return pipe_handler


@pytest.fixture
def pipe_handler_pipe():
    """With a working pipe!"""
    pipe_handler = PipeHandler(name="handler", context=FakeContext())  # type: ignore
    pipe_handler.internal_pipe = zmq.Context.instance().socket(zmq.PULL)
    pipe_handler.pipe_port = pipe_handler.internal_pipe.bind_to_random_port(
        "inproc://listenerPipe", min_port=12345)
    yield pipe_handler
    pipe_handler.close()


@pytest.fixture
def communicator(pipe_handler_pipe: PipeHandler) -> CommunicatorPipe:
    """Communicator of `pipe_handler_pipe`."""
    return pipe_handler_pipe.get_communicator()


def test_close_closes_all_communicators(
    pipe_handler_pipe: PipeHandler, communicator: CommunicatorPipe
):
    pipe_handler_pipe.close()
    assert communicator.socket.closed is True


class Test_PipeHandler_read_message:
    def test_handle_response(self, pipe_handler: PipeHandler):
        message = Message("rec", "send")
        pipe_handler.socket._r = [message.to_frames()]  # type: ignore
        pipe_handler.message_buffer.add_conversation_id(message.conversation_id)
        # act
        with pytest.raises(TimeoutError):
            pipe_handler.read_message()
        assert pipe_handler.message_buffer.wait_for_message(message.conversation_id) == message

    def test_handle_request(self, pipe_handler: PipeHandler, caplog: pytest.LogCaptureFixture):
        """Message is not a response and should be handled by the MessageHandler."""
        message = Message("rec", "send")
        pipe_handler.socket._r = [message.to_frames()]  # type: ignore
        # act and assert
        assert pipe_handler.read_message() == message


def test_invalid_pipe_message(pipe_handler: PipeHandler, caplog: pytest.LogCaptureFixture):
    caplog.set_level(10)
    pipe_handler.handle_pipe_message([b"abc"])
    assert caplog.messages[0].startswith("Received unknown")


class Test_get_communicator:
    @pytest.fixture
    def pipe_handler_setup(self):
        pipe_handler = PipeHandler(name="handler", context=FakeContext())  # type: ignore
        communicator = pipe_handler.get_communicator(context=FakeContext())  # type: ignore
        pipe_handler.external_pipe = communicator  # type: ignore
        return pipe_handler

    def test_external_pipe_type(self, pipe_handler_setup: PipeHandler):
        assert isinstance(pipe_handler_setup.external_pipe, CommunicatorPipe)  # type: ignore

    def test_pipe_ports_match(self, pipe_handler_setup: PipeHandler):
        port_number = pipe_handler_setup.pipe_port
        assert port_number == 5  # due to FakeSocket
        assert pipe_handler_setup.internal_pipe.addr == "inproc://listenerPipe"
        assert pipe_handler_setup.external_pipe.socket.addr == "inproc://listenerPipe:5"  # type: ignore  # noqa

    def test_second_call_returns_same_communicator(self, pipe_handler_setup: PipeHandler):
        com2 = pipe_handler_setup.get_communicator()
        assert com2 == pipe_handler_setup.external_pipe  # type: ignore


def test_close_all_communicators(pipe_handler_pipe: PipeHandler, communicator: CommunicatorPipe):
    pipe_handler_pipe.close_all_communicators()
    assert communicator.socket.closed is True


def test_communicator_send_message(pipe_handler_pipe: PipeHandler, communicator: CommunicatorPipe):
    message = Message("rec", "send")
    pipe_handler_pipe._send_frames = MagicMock()  # type: ignore[method-assign]
    communicator.send_message(message)
    pipe_handler_pipe.read_and_handle_pipe_message()
    # assert that the message is actually sent
    pipe_handler_pipe._send_frames.assert_called_once_with(frames=message.to_frames())


def test_communicator_send_message_without_sender(pipe_handler_pipe: PipeHandler,
                                                  communicator: CommunicatorPipe):
    message = Message("rec", sender="")
    pipe_handler_pipe._send_frames = MagicMock()  # type: ignore[method-assign]
    communicator.send_message(message)
    pipe_handler_pipe.read_and_handle_pipe_message()
    # assert that the message is actually sent
    message.sender = b"handler"  # should have been added by the handler
    pipe_handler_pipe._send_frames.assert_called_once_with(frames=message.to_frames())


def test_communicator_read_message(pipe_handler_pipe: PipeHandler, communicator: CommunicatorPipe):
    response = Message("handler", "rec", conversation_id=cid)
    pipe_handler_pipe.message_buffer.add_conversation_id(cid)
    pipe_handler_pipe.message_buffer.add_message(response)
    # act
    read = communicator.read_message(cid)
    assert read == response


def test_communicator_ask_message(pipe_handler_pipe: PipeHandler, communicator: CommunicatorPipe):
    message = Message("rec", "handler", conversation_id=cid)
    response = Message("handler", "rec", conversation_id=cid)
    pipe_handler_pipe._send_frames = MagicMock()  # type: ignore[method-assign]
    pipe_handler_pipe.message_buffer.add_conversation_id(cid)
    pipe_handler_pipe.message_buffer.add_message(response)
    # act
    read = communicator.ask_message(message)
    pipe_handler_pipe.read_and_handle_pipe_message()
    # assert
    assert read == response
    pipe_handler_pipe._send_frames.assert_called_once_with(frames=message.to_frames())


def test_communicator_sign_in(communicator: CommunicatorPipe):
    with pytest.raises(NotImplementedError):
        communicator.sign_in()


def test_communicator_sign_out(communicator: CommunicatorPipe):
    with pytest.raises(NotImplementedError):
        communicator.sign_out()


def test_communicator_subscribe(pipe_handler_pipe: PipeHandler, communicator: CommunicatorPipe):
    pipe_handler_pipe.subscribe_single = MagicMock()  # type: ignore[method-assign]
    # act
    communicator.subscribe_single(b"topic")
    pipe_handler_pipe.read_and_handle_pipe_message()
    # assert
    pipe_handler_pipe.subscribe_single.assert_called_once_with(topic=b"topic")


def test_communicator_unsubscribe(pipe_handler_pipe: PipeHandler, communicator: CommunicatorPipe):
    pipe_handler_pipe.unsubscribe_single = MagicMock()  # type: ignore[method-assign]
    # act
    communicator.unsubscribe_single(b"topic")
    pipe_handler_pipe.read_and_handle_pipe_message()
    # assert
    pipe_handler_pipe.unsubscribe_single.assert_called_once_with(topic=b"topic")


def test_communicator_unsubscribe_all(pipe_handler_pipe: PipeHandler,
                                      communicator: CommunicatorPipe):
    pipe_handler_pipe.unsubscribe_all = MagicMock()  # type: ignore[method-assign]
    # act
    communicator.unsubscribe_all()
    pipe_handler_pipe.read_and_handle_pipe_message()
    # assert
    pipe_handler_pipe.unsubscribe_all.assert_called_once()


def test_communicator_rename(pipe_handler_pipe: PipeHandler, communicator: CommunicatorPipe):
    pipe_handler_pipe.sign_in = MagicMock()  # type: ignore[method-assign]
    pipe_handler_pipe.sign_out = MagicMock()  # type: ignore[method-assign]
    # act
    communicator.name = "new name"
    pipe_handler_pipe.read_and_handle_pipe_message()
    # assert
    pipe_handler_pipe.sign_out.assert_called_once()
    assert pipe_handler_pipe.name == "new name"
    assert communicator.name == "new name"
    pipe_handler_pipe.sign_in.assert_called_once()


def test_register_name_change_method(pipe_handler: PipeHandler):
    method = MagicMock()
    pipe_handler.register_on_name_change_method(method)
    pipe_handler.set_full_name("new full name")
    # assert
    method.assert_called_once_with("new full name")


def test_unregister_name_change_method(pipe_handler: PipeHandler):
    method = MagicMock()
    pipe_handler.register_on_name_change_method(method)
    assert method in pipe_handler._on_name_change_methods
    # act
    pipe_handler.unregister_on_name_change_method(method)
    assert method not in pipe_handler._on_name_change_methods


def test_handle_local_method(pipe_handler_pipe: PipeHandler, communicator: CommunicatorPipe):
    cid = communicator._send_handler(method="pong")
    pipe_handler_pipe.read_and_handle_pipe_message()
    assert communicator._read_handler(cid) is None


def test_ask_handler(communicator: CommunicatorPipe):
    communicator._send_handler = MagicMock()  # type: ignore[method-assign]
    communicator._send_handler.return_value = b"conversation_id;"
    communicator._read_handler = MagicMock()  # type: ignore[method-assign]
    communicator._read_handler.return_value = 5
    # act
    result = communicator.ask_handler(method="method", timeout=1)
    assert result == 5
    communicator._send_handler.assert_called_once_with(method="method")
    communicator._read_handler.assert_called_once_with(b"conversation_id;", timeout=1)



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

import pytest

from pyleco.test import FakeContext
from pyleco.core.message import Message, MessageTypes
from pyleco.json_utils.errors import NOT_SIGNED_IN, DUPLICATE_NAME
from pyleco.json_utils.json_objects import Request, ResultResponse, ErrorResponse
from pyleco.utils.coordinator_utils import CommunicationError, ZmqNode, ZmqMultiSocket, Node,\
    Directory, FakeNode


class TestZmqMultiSocket:
    @pytest.fixture
    def socket(self):
        socket = ZmqMultiSocket(context=FakeContext())  # type: ignore
        socket._sock._r = [  # type: ignore
            [b"id", b"version", b"receiver", b"sender", b"header", b"data"]]
        return socket

    def test_poll_True(self, socket):
        assert socket.message_received() is True

    def test_read_message(self, socket):
        assert socket.read_message() == (b"id", Message.from_frames(
            b"version", b"receiver", b"sender", b"header", b"data"))

    def test_poll_False(self, socket):
        socket._sock._r = []
        assert socket.message_received() is False


class TestZmqNode:
    @pytest.fixture
    def node(self):
        node = ZmqNode(context=FakeContext())  # type: ignore
        return node

    def test_connection(self, node):
        node.connect("abc")
        assert node._dealer.addr == "tcp://abc"

    def test_disconnect(self, node):
        node.connect("abc")
        socket = node._dealer
        node.disconnect()
        assert socket.addr is None
        assert node.is_connected() is False

    def test_is_connected_with_dealer(self, node):
        node.connect("abc")
        node._dealer.close()
        assert node.is_connected() is False

    def test_is_connected(self, node):
        assert node.is_connected() is False
        node.connect("abc")
        assert node.is_connected() is True


cid = b"conversation_id;"


def fake_generate_cid():
    return cid


@pytest.fixture()
def fake_cid_generation(monkeypatch):
    monkeypatch.setattr("pyleco.core.serialization.generate_conversation_id", fake_generate_cid)


@pytest.fixture
def empty_directory() -> Directory:
    return Directory(b"N1", b"N1.COORDINATOR", "N1host:12300")


@pytest.fixture
def directory(empty_directory: Directory) -> Directory:
    d = empty_directory
    d._nodes[b"N2"] = n2 = FakeNode()
    d._node_ids[b"n2"] = n2
    n2.connect("N2host")
    n2.namespace = b"N2"
    d.add_component(b"send", b"send_id")
    d.add_component(b"rec", b"rec_id")
    return d


def fake_perf_counter():
    return 0.


@pytest.fixture()
def fake_counting(monkeypatch):
    monkeypatch.setattr("pyleco.utils.coordinator_utils.perf_counter", fake_perf_counter)


class Test_add_component:
    def test_add_component(self, empty_directory: Directory):
        empty_directory.add_component(b"name", b"identity")
        assert b"name" in empty_directory.get_components()
        assert empty_directory.get_components()[b"name"].identity == b"identity"

    def test_adding_component_of_already_signed_in_component_succeeds(self, fake_counting,
                                                                      directory: Directory):
        # TODO not defined in LECO
        sender = directory._components[b"send"]
        sender.heartbeat = -100
        directory.add_component(name=b"send", identity=b"send_id")
        # test that no error is raised.
        assert sender.heartbeat == 0

    def test_reject_adding(self, directory: Directory):
        with pytest.raises(ValueError):
            directory.add_component(b"send", b"new_identity")
        assert directory.get_components()[b"send"].identity != b"new_identity"


class Test_remove_component:
    @pytest.mark.parametrize("identity", (b"", b"send_id"))
    def test_remove(self, directory: Directory, identity):
        directory.remove_component(b"send", identity)
        assert b"send" not in directory.get_components()

    def test_remove_with_wrong_identity(self, directory: Directory):
        with pytest.raises(ValueError):
            directory.remove_component(b"send", b"wrong_identity")

    def test_remove_not_present(self, directory: Directory):
        """Test whether an already removed component does not raise an error."""
        directory.remove_component(b"not_present", b"")


class Test_get_component_id:
    def test_component_present(self, directory: Directory):
        assert directory.get_component_id(b"send") == b"send_id"

    def test_component_missing(self, directory: Directory):
        with pytest.raises(ValueError):
            directory.get_component_id(b"not_present")


class Test_add_node_sender:
    """These are the first two parts of the sign in process: connect and send the COSIGNIN."""
    @pytest.mark.parametrize("namespace", (b"N1", b"N2"))
    def test_invalid_namespaces(self, directory: Directory, namespace):
        with pytest.raises(ValueError):
            directory.add_node_sender(Node(), "N3host", namespace)

    @pytest.mark.parametrize("address", ("N1host", "N3host", "N1host:12300", "N3host:12300"))
    def test_invalid_address(self, directory: Directory, address):
        """No new connection to the coordinator itself or if there is another attempt to connect
        to that same remote node."""
        # simulate a connection to N3 under way
        directory._waiting_nodes["N3host:12300"] = None  # type: ignore
        with pytest.raises(ValueError):
            directory.add_node_sender(Node(), address, b"N3")

    def test_node_added(self, fake_counting, directory: Directory):
        length = len(directory._waiting_nodes)
        directory.add_node_sender(FakeNode(), "N3host", b"N3")
        assert directory._nodes.get(b"N3host") is None
        assert len(directory._waiting_nodes) > length

    @pytest.fixture
    def node(self, fake_cid_generation, directory: Directory) -> Node:
        address = "N3host:12345"
        directory.add_node_sender(node=FakeNode(), address=address, namespace=b"N3")
        return directory._waiting_nodes[address]

    def test_node_address(self, node: Node):
        assert node.address == "N3host:12345"

    def test_node_heartbeat(self, fake_counting, node: Node):
        assert node.heartbeat == 0

    def test_node_connected(self, node: Node):
        assert node.is_connected()

    def test_message_sent(self, node: Node):
        assert node._messages_sent == [Message(  # type: ignore
            b"COORDINATOR", b"N1.COORDINATOR",
            data=Request(id=1, method="coordinator_sign_in"),
            message_type=MessageTypes.JSON,
            conversation_id=cid)]

    def test_node_port_added_to_address(self, directory: Directory):
        directory.add_node_sender(FakeNode(), "N3host", b"N3")
        assert directory._waiting_nodes["N3host:12300"].address == "N3host:12300"


class Test_add_node_receiver_unknown:
    """Handles a remote Coordinator, which is signing in."""
    @pytest.fixture
    def unknown_node(self, fake_counting, directory: Directory) -> Node:
        identity = b"receiver_id"
        directory.add_node_receiver(identity=identity, namespace=b"N3")
        return directory._node_ids[identity]

    def test_namespace_with_identity(self, unknown_node: Node):
        assert unknown_node.namespace == b"N3"

    def test_heartbeat(self, unknown_node: Node):
        assert unknown_node.heartbeat == 0


class Test_add_node_receiver_partially_known:
    identity = b"n3"

    @pytest.fixture
    def partially_known_node(self, fake_counting, directory: Directory) -> Directory:
        """Only sending to that node is possible."""
        node = FakeNode()
        node.namespace = b"N3"
        directory._nodes[node.namespace] = node
        directory.add_node_receiver(identity=self.identity, namespace=b"N3")
        return directory

    def test_heartbeat(self, partially_known_node: Directory):
        assert partially_known_node._node_ids[self.identity].heartbeat == 0

    def test_id_set(self, partially_known_node: Directory):
        assert self.identity in partially_known_node._node_ids.keys()


def test_add_node_receiver_already_known(directory: Directory):
    with pytest.raises(ValueError):
        directory.add_node_receiver(b"n5", b"N2")


class Test_check_unfinished_node_connections:
    @pytest.fixture
    def directory_cunc(self, directory: Directory) -> Directory:
        directory.add_node_sender(FakeNode(), "N3host", b"N3")
        node = directory._waiting_nodes["N3host:12300"]
        node._messages_read = [Message(b"N1.COORDINATOR", b"N3.COORDINATOR",  # type: ignore
                                       message_type=MessageTypes.JSON,
                                       data=ResultResponse(id=1, result=None))]
        directory.check_unfinished_node_connections()
        return directory

    def test_new_node(self, directory_cunc: Directory):
        assert directory_cunc.get_node(b"N3") is not None


def test_check_unfinished_node_connection_logs_error(directory: Directory, caplog):
    directory.add_node_sender(FakeNode(), "N3host", b"N3")
    node = directory._waiting_nodes["N3host:12300"]

    def read_message(timeout: int = 0) -> Message:
        return Message.from_frames(*[b"frame 1", b"frame 2"])  # not enough frames
    node.read_message = read_message  # type: ignore
    node._messages_read = ["just something to indicate a message in the buffer"]  # type: ignore
    directory.check_unfinished_node_connections()
    assert caplog.records[-1].msg == "Message decoding failed."


class Test_handle_node_message:
    """Already included in check_unfinished_node_connections"""
    @pytest.mark.parametrize("message", (
            Message(b"N1.COORDINATOR", b"N5.COORDINATOR",
                    message_type=MessageTypes.JSON,
                    data=ErrorResponse(id=None, error=DUPLICATE_NAME)),
            Message(b"N1.COORDINATOR", b"N5.COORDINATOR",
                    message_type=MessageTypes.JSON,
                    data=ErrorResponse(id=None, error=NOT_SIGNED_IN)),
            Message("N1.COORDINATOR", "N5.COORDINATOR",
                    message_type=MessageTypes.JSON,
                    data={"jsonrpc": "2.0", "error": {"code": -32600, "message": "Invalid Request"},
                          "id": None})
    ))
    def test_rejected_sign_in(self, directory: Directory, message):
        directory._waiting_nodes["N5host"] = n = FakeNode()
        directory._handle_node_message(key="N5host", message=message)
        assert "N5host" not in directory._waiting_nodes.keys()
        assert n not in directory._nodes.values()

    @pytest.mark.parametrize("message", (
            Message(b"N1.COORDINATOR", b"N5.COORDINATOR",
                    message_type=MessageTypes.JSON,
                    data={"jsonrpc": "2.0", "result": None, "id": 1}),
            Message(b"N1.COORDINATOR", b"N5.COORDINATOR",
                    message_type=MessageTypes.JSON,
                    data={"jsonrpc": "2.0", "result": None, "id": 5}),
    ))
    def test_successful_sign_in(self, directory: Directory, message):
        directory._waiting_nodes["N5host"] = n = FakeNode()
        directory._handle_node_message(key="N5host", message=message)
        assert directory.get_node(b"N5") == n
        assert "N5host" not in directory._waiting_nodes.keys()


class Test_finish_sign_in_to_remote:
    @pytest.fixture
    def directory_sirn(self, fake_cid_generation, directory: Directory) -> Directory:
        directory.add_node_sender(FakeNode(), "N3host", b"N3")
        temp_namespace = list(directory._waiting_nodes.keys())[0]
        node = directory._waiting_nodes[temp_namespace]
        assert node.namespace != b"N3"
        directory._finish_sign_in_to_remote(temp_namespace, Message(
            receiver=b"N1.COORDINATOR",
            sender=b"N3.COORDINATOR",
            message_type=MessageTypes.JSON,
            data={"id": 1, "result": None, "jsonrpc": "2.0"},
        ))
        return directory

    def test_waiting_nodes_cleared(self, directory_sirn: Directory):
        assert len(directory_sirn._waiting_nodes) == 0

    def test_node_renamed(self, directory_sirn: Directory):
        assert list(directory_sirn._nodes.keys()) == [b"N2", b"N3"]

    def test_node_namespace_set(self, directory_sirn: Directory):
        assert directory_sirn._nodes[b"N3"].namespace == b"N3"

    def test_directory_sent(self, directory_sirn: Directory):
        assert directory_sirn._nodes[b"N3"]._messages_sent == [  # type: ignore
            Message(b'COORDINATOR', b'N1.COORDINATOR',
                    {"id": 1, "method": "coordinator_sign_in", "jsonrpc": "2.0"},
                    message_type=MessageTypes.JSON,
                    conversation_id=cid),
            Message(
                b'N3.COORDINATOR', b'N1.COORDINATOR',
                data=[{"id": 2, "method": "add_nodes", "params":
                       {"nodes": {"N1": "N1host:12300", "N2": "N2host", "N3": "N3host:12300"}},
                       "jsonrpc": "2.0"},
                      {"id": 3, "method": "record_components", "params":
                       {"components": ["send", "rec"]}, "jsonrpc": "2.0"}],
                message_type=MessageTypes.JSON,
                conversation_id=cid,
            )]


class Test_combine_sender_and_receiver_nodes:
    @pytest.fixture
    def directory_crasn(self, directory: Directory) -> Directory:
        directory.add_node_receiver(b"n3", b"N3")
        return directory

    def test_match(self, directory_crasn: Directory):
        node = Node()
        node.namespace = b"N3"
        directory_crasn._combine_sender_and_receiver_nodes(node)
        assert node.heartbeat != -1
        assert directory_crasn._node_ids[b"n3"] == node

    def test_mismatch(self, directory_crasn: Directory):
        node = Node()
        node.namespace = b"N4"
        directory_crasn._combine_sender_and_receiver_nodes(node)
        assert node.heartbeat == -1
        assert directory_crasn._node_ids[b"n3"] != node


class Test_remove_node_without_checks:
    def test_remove_only_receiver(self, directory: Directory):
        directory.add_node_receiver(b"n3", b"N3")
        assert b"n3" in directory._node_ids
        directory._remove_node_without_checks(b"N3")
        assert b"n3" not in directory._node_ids
        assert b"N3" not in directory._nodes

    def test_remove_complete_node(self, directory: Directory):
        assert b"n2" in directory._node_ids
        assert b"N2" in directory._nodes
        directory._remove_node_without_checks(b"N2")
        assert b"n2" not in directory._node_ids
        assert b"N2" not in directory._nodes

    def test_remove_by_namespace(self, directory: Directory):
        node = Node()
        directory._nodes[b"N3"] = node
        directory._node_ids[b"some_id"] = node
        directory._remove_node_without_checks(b"N3")
        assert b"some_id" not in directory._node_ids
        assert b"N3" not in directory._nodes


class Test_update_heartbeat:
    def test_local_component(self, fake_counting, directory: Directory):
        directory.update_heartbeat(b"send_id", Message.from_frames(b"", b"", b"N1.send", b""))
        assert directory.get_components()[b"send"].heartbeat == 0

    def test_local_component_without_namespace(self, fake_counting, directory: Directory):
        directory.update_heartbeat(b"send_id", Message.from_frames(b"", b"", b"send", b""))
        assert directory.get_components()[b"send"].heartbeat == 0

    def test_local_component_signs_in(self, directory: Directory):
        directory.update_heartbeat(b"new_id", Message.from_frames(
            b"", b"COORDINATOR", b"send2", b"",
            b'{"id": 2, "method": "sign_in", "jsonrpc": "2.0"}'))
        # test that no error is raised

    def test_not_signed_in_component_signs_out(self, directory: Directory):
        # TODO not determined by LECO
        directory.update_heartbeat(b"new_id", Message.from_frames(
            b"", b"COORDINATOR", b"send2", b"",
            b'{"id": 2, "method": "sign_out", "jsonrpc": "2.0"}'))
        # test that no error is raised

    def test_local_component_with_wrong_id(self, directory: Directory):
        with pytest.raises(CommunicationError, match=DUPLICATE_NAME.message):
            directory.update_heartbeat(b"new_id", Message.from_frames(
                b"", b"COORDINATOR", b"send", b""))

    def test_local_component_with_wrong_id_signs_in(self, directory: Directory):
        with pytest.raises(CommunicationError, match=DUPLICATE_NAME.message):
            directory.update_heartbeat(b"new_id", Message(
                receiver=b"COORDINATOR", sender=b"send",
                message_type=MessageTypes.JSON,
                data={"jsonrpc": "2.0", "id": 2, "method": "sign_in"}))

    def test_known_node(self, fake_counting, directory: Directory):
        directory.update_heartbeat(b"n2", Message.from_frames(b"", b"COORDINATOR", b"N2.send", b""))
        assert directory.get_node_ids()[b"n2"].heartbeat == 0

    @pytest.mark.parametrize("data", (
            {"jsonrpc": "2.0", "method": "coordinator_sign_in", "id": 7},
            {"jsonrpc": "2.0", "method": "coordinator_sign_out", "id": 7},
    ))
    def test_signing_in_out_node(self, directory: Directory, data):
        directory.update_heartbeat(b"n3", Message(
            b"COORDINATOR", b"N3.COORDINATOR", data=data, message_type=MessageTypes.JSON))
        # test that no error is raised

    def test_from_unknown_node(self, directory: Directory):
        with pytest.raises(CommunicationError, match="not signed in"):
            directory.update_heartbeat(b"n3", Message(b"COORDINATOR", b"N3.send"))


class Test_find_expired_components:
    def test_expired_component(self, directory: Directory, fake_counting):
        directory.get_components()[b"send"].heartbeat = -3.5
        directory.find_expired_components(1)
        assert b"send" not in directory.get_components().keys()

    def test_warn_component(self, directory: Directory, fake_counting):
        directory.get_components()[b"send"].heartbeat = -1.5
        assert directory.find_expired_components(1) == [(b"send_id", b"send")]

    def test_active_Component(self, directory: Directory, fake_counting):
        directory.get_components()[b"send"].heartbeat = -0.5
        assert directory.find_expired_components(1) == []
        assert b"send" in directory.get_components()


class Test_find_expired_nodes:

    def test_expired_node(self, directory: Directory, fake_counting):
        directory.get_node_ids()[b"n2"].heartbeat = -3.5
        directory.find_expired_nodes(1)
        assert b"n2" not in directory.get_node_ids()

    def test_warn_node(self, directory: Directory, fake_counting, fake_cid_generation):
        directory.get_node_ids()[b"n2"].heartbeat = -1.5
        directory.find_expired_nodes(1)
        assert directory.get_node_ids()[b"n2"]._messages_sent == [  # type: ignore
            Message(b"N2.COORDINATOR", b"N1.COORDINATOR",
                    Request(id=0, method="pong"),
                    message_type=MessageTypes.JSON,
                    conversation_id=cid)]

    def test_active_node(self, directory: Directory, fake_counting):
        directory.get_node_ids()[b"n2"].heartbeat = -0.5
        directory.find_expired_nodes(1)
        assert b"n2" in directory.get_node_ids()

    def test_expired_waiting_node(self, directory: Directory, fake_counting):
        waiting_node = FakeNode()
        waiting_node.heartbeat = - 3.5
        directory._waiting_nodes["address"] = waiting_node
        # act
        directory.find_expired_nodes(1)
        assert "address" not in directory._waiting_nodes


def test_get_node_id(directory: Directory):
    assert directory.get_node_id(b"N2") == b"n2"


def test_get_node_id_not_first_place_in_list(directory: Directory):
    n3 = FakeNode()
    n3.namespace = b"N3"
    directory._node_ids[b"n3"] = n3
    assert directory.get_node_id(b"N3") == b"n3"


def test_get_node_id_fails(directory: Directory):
    with pytest.raises(ValueError, match="No receiving connection to namespace"):
        directory.get_node_id(b"N5")


class Test_sign_out_from_node:
    @pytest.fixture
    def directory_wo_n2(self, fake_cid_generation, directory: Directory) -> Directory:
        directory._test = directory.get_node(b"N2")  # type: ignore
        directory.sign_out_from_node(b"N2")
        return directory

    def test_message_sent(self, directory_wo_n2: Directory):
        assert directory_wo_n2._test._messages_sent == [  # type: ignore
            Message(b"N2.COORDINATOR", b"N1.COORDINATOR",
                    data={"id": 1, "method": "coordinator_sign_out", "jsonrpc": "2.0"},
                    message_type=MessageTypes.JSON,
                    conversation_id=cid)]

    def test_connection_closed(self, directory_wo_n2: Directory):
        assert directory_wo_n2._test.is_connected() is False  # type: ignore

    def test_n2_removed_from_nodes(self, directory_wo_n2: Directory):
        assert b"N2" not in directory_wo_n2.get_nodes()

    def test_n2_removed_from_node_ids(self, directory_wo_n2: Directory):
        assert b"n2" not in directory_wo_n2.get_node_ids()


def test_sign_out_from_unknown_node_fails(directory: Directory):
    with pytest.raises(ValueError, match="is not known"):
        directory.sign_out_from_node(b"unknown node")



# -*- coding: utf-8 -*-
"""
Example scheme for an Actor for pymeasure instruments. 'pymeasure_actor'
"""
# This first docstring is shown in the GUI corresponding to the starter, such that it may be
# identified more easily.


import logging

from pyleco.actors.actor import Actor
from pyleco.utils.data_publisher import DataPublisher
from pymeasure.instruments.ipgphotonics import YAR  # type:ignore[import-not-found]

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


# Parameters
interval = 0.05  # Readout interval in s
adapter = "COM15"  # pymeasure adapter string


def readout(device: YAR, publisher: DataPublisher) -> None:
    """This method is executed every `interval`.

    :param device: The device driver managed by the Actor.
    :param publisher: The :class:`DataPublisher` instance of the Actor to publish data.
    """
    publisher.send_data(data={'power': device.power})


def task(stop_event) -> None:
    """The task which is run by the starter."""
    # Initialize
    with Actor(name="pymeasure_actor", device_class=YAR, periodic_reading=interval) as actor:
        actor.read_publish = readout  # define the regular readout function
        actor.connect(adapter)  # connect to the device

        # Continuous loop
        actor.listen(stop_event=stop_event)  # listen for commands and do the regular readouts

        # Finish
        # in listen and __exit__ included


if __name__ == "__main__":
    """Run the task if the module is executed."""
    log.addHandler(logging.StreamHandler())
    log.setLevel(logging.INFO)

    class Signal:
        def is_set(self):
            return False
    try:
        task(Signal())
    except KeyboardInterrupt:
        pass



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from __future__ import annotations
import logging
from typing import Generic, Optional, TypeVar, Union
from warnings import warn

from .director import Director


log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


Device = TypeVar("Device")


class RemoteCall:
    """Descriptor for remotely calling methods.

    You can add methods by simpling adding this Descriptor.
    Whenever this instance is called, it executes :code:`call_method`
    with the attribute name as `method` parameter. For example:

    .. code::

        class XYZ(BaseDirector):
            method = RemoteCall("Docstring for that method.")  # add a RemoteCall instance as attr.
        director = XYZ()
        director.method(*some_args, **kwargs)  # execute this instance.
        # equivalent to:
        director.call_method("method", *some_args, **kwargs)

    :param str name: Name of the method, only necessary if the RemoteCall is added after class
        creation.
    :param str doc: Docstring for the method. {name} is replaced by the attribute name of the
        instance of RemoteCall, in the example by 'method'.
    """

    def __init__(self, name: str = "", doc: Optional[str] = None, **kwargs) -> None:
        self._name = name
        if doc is None:
            doc = "Call '{name}' at the remote driver."
        self._doc = doc
        super().__init__(**kwargs)

    def __set_name__(self, owner, name) -> None:
        self._name = name
        self._doc = self._doc.format(name=self._name)

    def __get__(self, obj: Director, objtype=None):
        if obj is None:
            return self

        def remote_call(*args, **kwargs):
            obj.call_action(self._name, *args, **kwargs)

        remote_call.__doc__ = self._doc
        return remote_call


class TransparentDevice:
    """For all property access, the remote device is called.

    If you want to call methods, you can add them. with :class:`RemoteCall` to a subclass of this
    instrument.
    """

    director: Director

    def __init__(self, director: Director):
        self.director = director

    def call_action(self, action: str, *args, **kwargs):
        self.director.call_action(action, *args, **kwargs)

    def __getattr__(self, name):
        if name in dir(self):
            return super().__getattribute__(name)
        else:
            return self.director.get_parameters(parameters=(name,)).get(name)

    def __setattr__(self, name, value) -> None:
        if name in dir(self) or name.startswith("_") or name in ("director"):
            super().__setattr__(name, value)
        else:
            self.director.set_parameters(parameters={name: value})

    # TODO generate a list of capabilities of the actor and return these capabilities during a call
    # to __dir__. That enables autocompletion etc.


class TransparentDirector(Director, Generic[Device]):
    """Director getting/setting all properties remotely.

    It has a :attr:`device` attribute. Whenever you get/set an attribute of `device`, the Director
    will call the Actor and try to get/set the corresponding attribute of the Actor's device.
    If you want to add method calls, you might use the :class:`RemoteCall` Descriptor to add methods
    to a subclass of :class:`TransparentDevice` and give that class to the `device_class` parameter.
    For example :code:`method = RemoteCall()` in the class definition will make sure,
    that :code:`device.method(*args, **kwargs)` will be executed remotely.

    :param actor: Name of the actor to direct.
    :param device_class: Subclass of :class:`TransparentDevice` to use as a device dummy.
    :param cls: see :code:`device_class`.

        .. deprecated:: 0.3
            Use :code:`device_class` instead.
    """

    def __init__(
        self,
        actor: Optional[Union[bytes, str]] = None,
        device_class: type[Device] = TransparentDevice,  # type: ignore[assignment]
        cls: Optional[type[Device]] = None,
        **kwargs,
    ):
        super().__init__(actor=actor, **kwargs)
        if cls is not None:
            warn("Parameter `cls` is deprecated, use `device_class` instead.", FutureWarning)
            device_class = cls
        self.device = device_class(director=self)  # type: ignore[call-arg]



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from __future__ import annotations
from json import JSONDecodeError
from typing import Any, Iterable, Optional, Union


from . import VERSION_B
from .serialization import (create_header_frame, serialize_data, interpret_header, split_name,
                            deserialize_data, FullName, Header, MessageTypes,
                            )


# Control transfer protocol
class Message:
    """A message in LECO protocol.

    It consists of the following bytes frames, for which each an attribute exists:
        - `version`
        - `receiver`
        - `sender`
        - `header`
        - 0 or more `payload` frames

    If you do not specify a sender, the sending program shall add it itself.
    The :attr:`data` attribute is the content of the first :attr:`payload` frame. It can be set with
    the corresponding argument.
    All attributes, except the official frames, are for convenience.
    """

    version: bytes = VERSION_B
    receiver: bytes
    sender: bytes
    header: bytes
    payload: list[bytes]

    def __init__(self,
                 receiver: Union[bytes, str],
                 sender: Union[bytes, str] = b"",
                 data: Optional[Union[bytes, str, Any]] = None,
                 header: Optional[bytes] = None,
                 conversation_id: Optional[bytes] = None,
                 message_id: Optional[bytes] = None,
                 message_type: Union[MessageTypes, int] = MessageTypes.NOT_DEFINED,
                 additional_payload: Optional[Iterable[bytes]] = None,
                 ) -> None:
        self.receiver = receiver.encode() if isinstance(receiver, str) else receiver
        self.sender = sender.encode() if isinstance(sender, str) else sender
        if header and (conversation_id or message_id or message_type != MessageTypes.NOT_DEFINED):
            raise ValueError(
                "You may not specify the header and some header element at the same time!")
        self.header = (create_header_frame(conversation_id=conversation_id, message_id=message_id,
                                           message_type=message_type)
                       if header is None else header)
        if isinstance(data, bytes):
            self.payload = [data]
        elif isinstance(data, str):
            self.payload = [data.encode()]
        elif data is None:
            self.payload = []
        else:
            self.payload = [serialize_data(data)]
        if additional_payload is not None:
            self.payload.extend(additional_payload)

    @classmethod
    def from_frames(cls, version: bytes, receiver: bytes, sender: bytes, header: bytes,
                    *payload: bytes):  # -> typing.Self for py>=3.11
        """Create a message from a frames list, for example after reading from a socket.

        .. code::

            frames = socket.recv_multipart()
            message = Message.from_frames(*frames)
        """
        inst = cls(receiver, sender, header=header, additional_payload=payload)
        inst.version = version
        return inst

    def to_frames(self) -> list[bytes]:
        """Get a list representation of the message, ready for sending it."""
        if not self.sender:
            raise ValueError("Empty sender frame not allowed to send.")
        return self._to_frames_without_sender_check()

    def _to_frames_without_sender_check(self) -> list[bytes]:
        return [self.version, self.receiver, self.sender, self.header] + self.payload

    # Convenience methods to access elements
    @property
    def receiver_elements(self) -> FullName:
        return split_name(self.receiver)

    @property
    def sender_elements(self) -> FullName:
        return split_name(self.sender)

    @property
    def header_elements(self) -> Header:
        return interpret_header(self.header)

    @property
    def conversation_id(self) -> bytes:
        return self.header_elements.conversation_id

    @property
    def data(self) -> object:
        return deserialize_data(self.payload[0]) if self.payload else None

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Message):
            return NotImplemented
        partial_comparison = (
            self.version == other.version
            and self.receiver == other.receiver
            and self.sender == other.sender
            and self.header == other.header
        )
        try:
            # Try to compare the data (python objects) instead of their bytes representation.
            my_data = self.data
            other_data = other.data
        except JSONDecodeError:
            # Maybe the payload is binary, compare the raw payload
            return partial_comparison and self.payload == other.payload
        else:
            return (partial_comparison and my_data == other_data
                    and self.payload[1:] == other.payload[1:])

    def __repr__(self) -> str:
        list_of_frames_strings = [
            str(frame) for frame in self._to_frames_without_sender_check()]
        return f"Message.from_frames({', '.join(list_of_frames_strings)})"



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

import pytest

from pyleco.test import FakeCommunicator
from pyleco.core.message import Message

from pyleco.utils.listener import Listener, CommunicatorPipe


@pytest.fixture
def listener() -> Listener:
    listener = Listener(name="test")  # type: ignore
    listener.communicator = FakeCommunicator(name="N.Pipe")  # type: ignore
    return listener


def test_communicator_name_is_returned(listener: Listener):
    assert listener.name == "N.Pipe"


class Test_communicator_closed_at_stopped_listener():
    @pytest.fixture(scope="class")
    def communicator(self) -> CommunicatorPipe:
        # scope is class as starting the listener takes some time
        listener = Listener(name="test")
        listener.start_listen()
        communicator = listener.communicator
        listener.stop_listen()
        return communicator

    def test_socket_closed(self, communicator: CommunicatorPipe):
        assert communicator.socket.closed is True

    def test_internal_method(self, communicator: CommunicatorPipe):
        """A method which is handled in the handler and not sent from the handler via LECO."""
        with pytest.raises(ConnectionRefusedError):
            communicator.ask_handler("pong")

    def test_sending_messages(self, communicator: CommunicatorPipe):
        with pytest.raises(ConnectionRefusedError):
            communicator.send_message(Message("rec", "send"))

    def test_changing_name(self, communicator: CommunicatorPipe):
        with pytest.raises(ConnectionRefusedError):
            communicator.name = "abc"



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from typing import Optional

from qtpy.QtCore import QObject, Signal  # type: ignore
from zmq import Context  # type: ignore

from ..core.message import Message
from ..core.data_message import DataMessage
from .listener import Listener, PipeHandler


class ListenerSignals(QObject):
    """Signals for the Listener."""
    # General
    name_changed = Signal(str)
    # Control protocol
    json_request_message = Signal(Message)
    json_error_message = Signal(Message)
    json_result_message = Signal(Message)
    message = Signal(Message)  # emitted in the same cases as above messages.
    # Data Protocol
    dataReady = Signal(dict)
    data_message = Signal(DataMessage)


class QtPipeHandler(PipeHandler):

    local_methods = ["pong", "set_log_level"]

    def __init__(self, name: str, signals: ListenerSignals, context: Optional[Context] = None,
                 **kwargs) -> None:
        self.signals = signals
        super().__init__(name, context, **kwargs)

    def handle_json_request(self, message: Message) -> None:
        try:
            method = message.data.get("method")  # type: ignore
        except AttributeError:
            pass
        else:
            if method in self.local_methods:
                super().handle_json_request(message=message)
                return
        # in all other cases:
        self.signals.message.emit(message)
        self.signals.json_request_message.emit(message)

    def handle_json_error(self, message: Message) -> None:
        self.signals.message.emit(message)
        self.signals.json_error_message.emit(message)

    def handle_json_result(self, message: Message) -> None:
        self.signals.message.emit(message)
        self.signals.json_result_message.emit(message)

    def handle_unknown_message_type(self, message: Message) -> None:
        self.signals.message.emit(message)

    def handle_subscription_data(self, data: dict) -> None:
        """Handle incoming subscription data."""
        # old style
        self.signals.dataReady.emit(data)

    def handle_subscription_message(self, message: DataMessage) -> None:
        """Handle an incoming subscription message."""
        # new style
        self.signals.data_message.emit(message)


class QtListener(Listener):
    """Listening on incoming messages in a separate thread - PyQt version.

    On one side it handles incoming messages (data and control protocol) in another thread.
    On the other side, it offers the :meth:`get_communicator` method, which returns a
    :class:`Communicator`, offering communication to the network.

    Call :meth:`.start_listen()` to actually listen.

    It emits signals from :attr:`signals` if a control or data message arrives.
    It also emits the `signals.name_changed` signal, whenever the Communicator changes its name.

    :param int data_port: Configure the port to be used for configuration.
    :param logger: Logger instance whose logs should be published. Defaults to "__main__".
    """

    def __init__(self, name: str, host: str = "localhost", **kwargs) -> None:
        super().__init__(name=name, host=host, **kwargs)
        self.signals = ListenerSignals()

    def _listen(self, name: str, stop_event, coordinator_host: str, coordinator_port: int,
                data_host: str, data_port: int) -> None:
        self.message_handler = QtPipeHandler(name, signals=self.signals,
                                             host=coordinator_host, port=coordinator_port,
                                             data_host=data_host, data_port=data_port,)
        self.message_handler.register_on_name_change_method(self.signals.name_changed.emit)
        self.message_handler.listen(stop_event=stop_event)



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

import pytest

from pyleco.core.data_message import DataMessage

topic = b"N1.CA"
cid = b"conversation_id;"
message_type = 0


@pytest.fixture
def message():
    message = DataMessage(topic, conversation_id=cid, data=[4, 5])
    return message


class TestDataMessageInit:
    def test_topic(self, message: DataMessage):
        assert message.topic == topic

    def test_header(self, message: DataMessage):
        assert message.header == cid + b"\x00"

    def test_data(self, message: DataMessage):
        assert message.data == [4, 5]

    @pytest.mark.parametrize("key, value", (("conversation_id", b"content"),
                                            ("message_type", 7),
                                            ))
    def test_header_param_incompatible_with_header_element_params(self, key, value):
        with pytest.raises(ValueError, match="header"):
            DataMessage(topic="topic", header=b"whatever", **{key: value})

    def test_additional_payload(self):
        message = DataMessage("topic", data=b"0", additional_payload=[b"1", b"2"])
        assert message.payload == [b"0", b"1", b"2"]

    def test_additional_payload_without_data(self):
        message = DataMessage("topic", additional_payload=[b"1", b"2"])
        assert message.payload == [b"1", b"2"]


def test_data_message_str_topic():
    assert DataMessage(topic="topic").topic == b"topic"


def test_data_message_message_type():
    message = DataMessage(topic=b"topic", message_type=7)
    assert message.message_type == 7


def test_converation_id(message: DataMessage):
    assert message.conversation_id == cid


def test_message_type(message: DataMessage):
    assert message.message_type == 0


class TestDataMessageData:
    def test_bytes_payload(self):
        message = DataMessage(b"", data=b"some data stuff")
        assert message.payload == [b"some data stuff"]

    def test_str_payload(self):
        message = DataMessage(b"", data="some data stuff")
        assert message.payload == [b"some data stuff"]

    def test_no_payload(self):
        message = DataMessage(b"")
        assert message.payload == []


class TestFromFrames:
    @pytest.fixture
    def from_frames_message(self):
        message = DataMessage.from_frames(b"frame0", b"frame1", b"frame2", b"frame3")
        return message

    def test_topic(self, from_frames_message: DataMessage):
        assert from_frames_message.topic == b"frame0"

    def test_header(self, from_frames_message: DataMessage):
        assert from_frames_message.header == b"frame1"

    def test_payload(self, from_frames_message: DataMessage):
        assert from_frames_message.payload == [b"frame2", b"frame3"]


def test_to_frames():
    message = DataMessage(b"topic", conversation_id=cid, data=b"data")
    assert message.to_frames() == [b"topic", b"conversation_id;\x00", b"data"]


class TestComparison:
    def test_message_comparison(self):
        frames = [b"topic", b"conversation_id;\x00", b'[["GET", [1, 2]]']
        m1 = DataMessage.from_frames(*frames)
        m2 = DataMessage.from_frames(*frames)
        assert m1 == m2

    def test_dictionary_order_is_irrelevant(self):
        m1 = DataMessage(b"topic", conversation_id=cid, data={"a": 1, "b": 2})
        m2 = DataMessage(b"topic", conversation_id=cid, data={"b": 2, "a": 1})
        assert m1 == m2

    def test_distinguish_empty_payload_frame(self):
        m1 = DataMessage("r", conversation_id=b"conversation_id;")
        m1.payload = [b""]
        m2 = DataMessage("r", conversation_id=b"conversation_id;")
        assert m2.payload == []  # verify that it does not have a payload
        assert m1 != m2

    @pytest.mark.parametrize("other", (5, 3.4, [64, 3], (5, "string"), "string"))
    def test_comparison_with_something_else_fails(self, message, other):
        assert message != other


def test_repr():
    message = DataMessage.from_frames(b'topic', b'conversation_id;\x00', b'data')
    assert repr(message) == r"DataMessage.from_frames(b'topic', b'conversation_id;\x00', b'data')"



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from __future__ import annotations
import datetime
import uuid
from typing import Any, Optional, Union

import pytest
from pyleco.json_utils.json_objects import Request

from pyleco.core import serialization
from pyleco.core.serialization import JsonContentTypes, get_json_content_type


class Test_create_header_frame:
    @pytest.mark.parametrize("kwargs, header", (
        ({"conversation_id": b"\x00" * 16, }, bytes([0] * 20)),
        ({"conversation_id": b"\x00" * 16, "message_type": b"5"}, bytes([0] * 19) + b"5")
    ))
    def test_create_header_frame(self, kwargs, header):
        assert serialization.create_header_frame(**kwargs) == header

    @pytest.mark.parametrize("cid_l", (0, 1, 2, 4, 6, 7, 10, 15, 17, 20))
    def test_wrong_cid_length_raises_errors(self, cid_l):
        with pytest.raises(ValueError, match="'conversation_id'"):
            serialization.create_header_frame(conversation_id=bytes([0] * cid_l))

    @pytest.mark.parametrize("mid_l", (0, 1, 2, 4, 6, 7))
    def test_wrong_mid_length_raises_errors(self, mid_l):
        with pytest.raises(ValueError, match="'message_id'"):
            serialization.create_header_frame(message_id=bytes([0] * mid_l))

    @pytest.mark.parametrize("mtl", (0, 2, 3))
    def test_wrong_m_type_length_raises_errors(self, mtl):
        with pytest.raises(ValueError, match="'message_type'"):
            serialization.create_header_frame(message_type=bytes([0] * mtl))


@pytest.mark.parametrize("header, conversation_id, message_id, message_type", (
        (bytes(range(20)), bytes(range(16)), b"\x10\x11\x12", 19),
))
def test_interpret_header(header, conversation_id, message_id, message_type):
    assert serialization.interpret_header(header) == (conversation_id, message_id, message_type)


@pytest.mark.parametrize("full_name, node, name", (
    (b"local only", b"node", b"local only"),
    (b"abc.def", b"abc", b"def"),
))
def test_split_name(full_name, node, name):
    assert serialization.split_name(full_name, b"node") == (node, name)


@pytest.mark.parametrize("full_name, node, name", (
    ("local only", "node", "local only"),
    ("abc.def", "abc", "def"),
))
def test_split_name_str(full_name, node, name):
    assert serialization.split_name_str(full_name, "node") == (node, name)


class Test_serialize:
    def test_json_object(self):
        obj = Request(id=3, method="whatever")
        expected = b'{"id":3,"method":"whatever","jsonrpc":"2.0"}'
        assert serialization.serialize_data(obj) == expected

    def test_dict(self):
        raw = {"some": "item", "key": "value", 5: [7, 3.1]}
        expected = b'{"some":"item","key":"value","5":[7,3.1]}'
        assert serialization.serialize_data(raw) == expected


class Test_generate_conversation_id_is_UUIDv7:
    @pytest.fixture
    def conversation_id(self):
        return serialization.generate_conversation_id()

    def test_type_is_bytes(self, conversation_id):
        assert isinstance(conversation_id, bytes)

    def test_length(self, conversation_id):
        assert len(conversation_id) == 16

    def test_UUID_version(self, conversation_id):
        assert conversation_id[6] >> 4 == 7

    def test_variant(self, conversation_id):
        assert conversation_id[8] >> 6 == 0b10

    def test_correct_timestamp(self, conversation_id):
        ts = serialization.conversation_id_to_datetime(conversation_id=conversation_id)
        assert abs(ts - datetime.datetime.now(datetime.timezone.utc)) < datetime.timedelta(hours=1)


def test_conversation_id_to_datetime_according_to_uuid_example():
    """According to the draft https://datatracker.ietf.org/doc/draft-ietf-uuidrev-rfc4122bis/14/
    017F22E2-79B0-7CC3-98C4-DC0C0C07398F should be a timestamp of
    Tuesday, February 22, 2022 2:22:22.00 PM GMT-05:00, represented as 1645557742000
    """
    cid = uuid.UUID("017F22E2-79B0-7CC3-98C4-DC0C0C07398F").bytes
    reference = datetime.datetime(
        2022, 2, 22, 14, 22, 22, tzinfo=datetime.timezone(datetime.timedelta(hours=-5))
    )
    ts = serialization.conversation_id_to_datetime(cid)
    assert reference - ts == datetime.timedelta(0)


def test_json_type_result_is_response():
    assert JsonContentTypes.RESPONSE in JsonContentTypes.RESULT_RESPONSE
    assert JsonContentTypes.RESULT in JsonContentTypes.RESULT_RESPONSE


def test_json_type_error_is_response():
    assert JsonContentTypes.RESPONSE in JsonContentTypes.ERROR_RESPONSE
    assert JsonContentTypes.ERROR in JsonContentTypes.ERROR_RESPONSE


# Methods for get_json_content_type
def create_request(method: str, params: Optional[Union[list, dict]] = None, id: int = 1
                   ) -> dict[str, Any]:
    return {"jsonrpc": "2.0", "id": id, "method": method, "params": params}


def create_result(result: Any, id: int = 1) -> dict[str, Any]:
    return {"jsonrpc": "2.0", "result": result, "id": id}


def create_error(error_code: int, error_message: str, id: int = 1) -> dict[str, Any]:
    return {"jsonrpc": "2.0", "id": id, "error": {"code": error_code, "message": error_message}}


class Test_get_json_content_type:
    @pytest.mark.parametrize("data, type", (
            (create_request("abc"), JsonContentTypes.REQUEST),
            ([create_request(method="abc")] * 2, JsonContentTypes.REQUEST | JsonContentTypes.BATCH),
            (create_result(None), JsonContentTypes.RESULT_RESPONSE),
            ([create_result(None), create_result(5, 7)],
             JsonContentTypes.RESULT_RESPONSE | JsonContentTypes.BATCH),
            (create_error(89, "whatever"), JsonContentTypes.ERROR_RESPONSE),
            ([create_error(89, "xy")] * 2,
             JsonContentTypes.ERROR_RESPONSE | JsonContentTypes.BATCH),
            ([create_result(4), create_error(32, "xy")],  # batch of result and error
             JsonContentTypes.RESULT_RESPONSE | JsonContentTypes.BATCH | JsonContentTypes.ERROR),
    ))
    def test_data_is_valid_type(self, data, type):
        assert get_json_content_type(data) == type

    @pytest.mark.parametrize("data", (
        {},
        [],
        [{}],
        {"some": "thing"},
        5.6,
        "adsfasdf",
    ))
    def test_invalid_data(self, data):
        assert get_json_content_type(data) == JsonContentTypes.INVALID



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

import pytest

from pyleco.core import VERSION_B
from pyleco.core.serialization import serialize_data

from pyleco.core.message import Message, MessageTypes


cid = b"conversation_id;"


@pytest.fixture
def message() -> Message:
    return Message(
        receiver=b"N1.receiver",
        sender=b"N2.sender",
        data=[["GET", [1, 2]], ["GET", 3]],
        conversation_id=cid,
        message_id=b"mid",
        message_type=int.from_bytes(b"T", byteorder="big")
    )


class Test_Message_create_message:
    def test_payload(self, message: Message):
        assert message.payload == [serialize_data([["GET", [1, 2]], ["GET", 3]])]

    def test_version(self, message: Message):
        assert message.version == VERSION_B

    def test_receiver(self, message: Message):
        assert message.receiver == b"N1.receiver"

    def test_sender(self, message: Message):
        assert message.sender == b"N2.sender"

    def test_header(self, message: Message):
        assert message.header == b"conversation_id;midT"

    def test_to_frames(self, message: Message):
        assert message.to_frames() == [
            VERSION_B,
            b"N1.receiver",
            b"N2.sender",
            b"conversation_id;midT",
            serialize_data([["GET", [1, 2]], ["GET", 3]]),
        ]

    def test_message_without_data_does_not_have_payload_frame(self):
        message = Message(b"N1.receiver", b"N2.sender", conversation_id=b"conversation_id;")
        assert message.payload == []
        assert message.to_frames() == [VERSION_B, b"N1.receiver", b"N2.sender",
                                       b"conversation_id;\x00\x00\x00\x00"]

    def test_message_binary_data(self):
        message = Message(b"N1.receiver", data=b"binary data")
        assert message.payload[0] == b"binary data"

    def test_message_data_str_to_binary_data(self):
        message = Message(b"rec", data="some string")
        assert message.payload[0] == b"some string"

    def test_additional_binary_data(self):
        message = Message(b"rec", data=b"0", additional_payload=[b"1", b"2"])
        assert message.payload == [b"0", b"1", b"2"]

    def test_additional_payload_without_data(self):
        message = Message(b"rec", additional_payload=[b"1", b"2"])
        assert message.payload == [b"1", b"2"]

    @pytest.mark.parametrize("key, value", (("conversation_id", b"content"),
                                            ("message_id", b"mid"),
                                            ("message_type", 7),
                                            ))
    def test_header_param_incompatible_with_header_element_params(self, key, value):
        with pytest.raises(ValueError, match="header"):
            Message(receiver=b"", header=b"whatever", **{key: value})


class Test_Message_from_frames:
    def test_message_from_frames(self, message: Message):
        message.version = b"diff"  # also if the version is different
        assert Message.from_frames(*message.to_frames()) == message

    def test_different_version(self, message: Message):
        message.version = b"diff"  # also if the version is different
        new_message = Message.from_frames(*message.to_frames())
        assert new_message.version == b"diff"

    def test_multiple_payload_frames(self):
        message = Message.from_frames(
            b"\xffo", b"broker", b"", b";", b'[["GET", [1, 2]], ["GET", 3]]', b"additional stuff"
        )
        assert message.payload == [b'[["GET", [1, 2]], ["GET", 3]]', b"additional stuff"]

    def test_no_payload(self):
        message = Message.from_frames(VERSION_B, b"broker", b"", b"")
        assert message.payload == []


def test_to_frames_without_payload(message: Message):
    message.payload = []
    assert message.to_frames() == [VERSION_B, b"N1.receiver", b"N2.sender",
                                   b"conversation_id;midT"]


class Test_Message_frame_splitting:
    """Test whether the splitting of header/sender/receiver works as expected."""

    def test_receiver_namespace(self, message: Message):
        assert message.receiver_elements.namespace == b"N1"

    def test_receiver_name(self, message: Message):
        assert message.receiver_elements.name == b"receiver"

    def test_sender_namespace(self, message: Message):
        assert message.sender_elements.namespace == b"N2"

    def test_sender_name(self, message: Message):
        assert message.sender_elements.name == b"sender"

    def test_header_conversation_id(self, message: Message):
        assert message.header_elements.conversation_id == cid

    def test_header_message_id(self, message: Message):
        assert message.header_elements.message_id == b"mid"

    def test_header_message_type(self, message: Message):
        assert message.header_elements.message_type == int.from_bytes(b"T", byteorder="big")


class Test_Message_with_string_parameters:
    @pytest.fixture
    def str_message(self) -> Message:
        message = Message(receiver="N2.receiver", sender="N1.sender")
        return message

    def test_receiver_is_bytes(self, str_message: Message):
        assert str_message.receiver == b"N2.receiver"

    def test_sender_is_bytes(self, str_message: Message):
        assert str_message.sender == b"N1.sender"


class Test_Message_data_payload_conversion:
    def test_data_to_payload(self):
        message = Message(b"r", b"s", data=([{5: "1asfd"}], 8), message_type=MessageTypes.JSON)
        assert message.payload == [serialize_data([[{"5": "1asfd"}], 8])]
        assert message.data == [[{'5': "1asfd"}], 8]  # converted to and from json, so modified!

    def test_payload_to_data(self):
        frames = [b"v", b"r", b"s", b"h", b'[["G", ["nodes"]]]', b'p2']
        message = Message.from_frames(*frames)
        assert message.payload == [b'[["G", ["nodes"]]]', b'p2']
        assert message.data == [["G", ["nodes"]]]

    def test_no_payload_is_no_data(self):
        message = Message(b"r")
        message.payload = []  # make sure, that there is no payload
        assert message.data is None


def test_get_frames_list_raises_error_on_empty_sender():
    m = Message(b"r")
    with pytest.raises(ValueError):
        m.to_frames()


class TestComparison:
    def test_message_comparison(self):
        frames = [VERSION_B, b"rec", b"send", b"x;y", b'[["GET", [1, 2]]']
        m1 = Message.from_frames(*frames)
        m2 = Message.from_frames(*frames)
        assert m1 == m2

    def test_dictionary_order_is_irrelevant(self):
        m1 = Message(b"r", conversation_id=cid, data={"a": 1, "b": 2},
                     message_type=MessageTypes.JSON)
        m2 = Message(b"r", conversation_id=cid, data={"b": 2, "a": 1},
                     message_type=MessageTypes.JSON)
        assert m1 == m2

    def test_distinguish_empty_payload_frame(self):
        m1 = Message("r", conversation_id=b"conversation_id;")
        m1.payload = [b""]
        m2 = Message("r", conversation_id=b"conversation_id;")
        assert m2.payload == []  # verify that it does not have a payload
        assert m1 != m2

    @pytest.mark.parametrize("other", (5, 3.4, [64, 3], (5, "string"), "string"))
    def test_comparison_with_something_else_fails(self, message, other):
        assert message != other


def test_repr():
    message = Message.from_frames(b'V', b'rec', b'send', b'cid;mid', b'data')
    assert repr(message) == r"Message.from_frames(b'V', b'rec', b'send', b'cid;mid', b'data')"


def test_repr_without_sender():
    message = Message.from_frames(b'V', b'rec', b'', b'cid;mid', b'data')
    assert repr(message) == r"Message.from_frames(b'V', b'rec', b'', b'cid;mid', b'data')"


def test_conversation_id_getter(message: Message):
    assert message.conversation_id == cid



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from __future__ import annotations
import logging

from .director import Director


log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


class CoordinatorDirector(Director):
    """Direct a Coordinator."""

    def __init__(self, actor="COORDINATOR", **kwargs) -> None:
        super().__init__(actor=actor, **kwargs)

    def get_local_components(self) -> list[str]:
        """Get the directory."""
        return self.ask_rpc(method="send_local_components")

    def get_global_components(self) -> dict[str, list[str]]:
        """Get the directory."""
        return self.ask_rpc(method="send_global_components")

    def get_nodes(self) -> dict[str, str]:
        """Get all known nodes."""
        return self.ask_rpc(method="send_nodes")

    def add_nodes(self, coordinators: dict[str, str]) -> None:
        """Tell the Coordinator about other coordinators (dict)."""
        return self.ask_rpc(method="add_nodes", nodes=coordinators)



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from __future__ import annotations
import datetime
from enum import IntEnum, IntFlag
import json
from typing import Any, Optional, NamedTuple, Union

# as long as uuid does not yet support UUIDv7 use uuid6
from uuid6 import uuid7
from ..json_utils.json_objects import (
    Request,
    ParamsRequest,
    ResultResponse,
    ErrorResponse,
    Notification,
    ParamsNotification,
)


json_objects = (
    Request,
    ParamsRequest,
    ResultResponse,
    ErrorResponse,
    Notification,
    ParamsNotification,
)


class FullName(NamedTuple):
    namespace: bytes
    name: bytes


class FullNameStr(NamedTuple):
    namespace: str
    name: str


class Header(NamedTuple):
    conversation_id: bytes
    message_id: bytes
    message_type: int


class MessageTypes(IntEnum):
    """The different message types, represented as an integer in the range [0, 255]."""
    NOT_DEFINED = 0
    JSON = 1


class JsonContentTypes(IntFlag):
    """Type of the JSON content."""
    INVALID = 0
    REQUEST = 1
    RESPONSE = 2
    RESULT = 4
    ERROR = 8
    BATCH = 16
    RESULT_RESPONSE = RESPONSE + RESULT
    ERROR_RESPONSE = RESPONSE + ERROR


def create_header_frame(conversation_id: Optional[bytes] = None,
                        message_id: Optional[Union[bytes, int]] = 0,
                        message_type: Union[bytes, int, MessageTypes] = MessageTypes.NOT_DEFINED,
                        ) -> bytes:
    """Create the header frame.

    :param bytes conversation_id: ID of the conversation.
    :param bytes message_id: Message ID of this message.
    :return: header frame.
    """
    if conversation_id is None:
        conversation_id = generate_conversation_id()
    elif (length := len(conversation_id)) != 16:
        raise ValueError(f"Length of 'conversation_id' is {length}, not 16 bytes.")
    if message_id is None:
        message_id = b"\x00" * 3
    elif isinstance(message_id, int):
        message_id = message_id.to_bytes(length=3, byteorder='big')
    elif len(message_id) != 3:
        raise ValueError("Length of 'message_id' is not 3 bytes.")
    if isinstance(message_type, int):
        message_type = message_type.to_bytes(length=1, byteorder="big")
    elif len(message_type) != 1:
        raise ValueError("Length of 'message_type' is not 1 bytes.")
    return b"".join((conversation_id, message_id, message_type))


def split_name(name: bytes, namespace: bytes = b"") -> FullName:
    """Split a sender/receiver name with given default namespace."""
    s = name.split(b".")
    n = s.pop(-1)
    return FullName((s.pop() if s else namespace), n)


def split_name_str(name: str, namespace: str = "") -> FullNameStr:
    """Split a sender/receiver name with given default namespace."""
    s = name.split(".")
    n = s.pop(-1)
    return FullNameStr((s.pop() if s else namespace), n)


def interpret_header(header: bytes) -> Header:
    """Interpret the header frame.

    :return: conversation_id, message_id, message_type
    """
    conversation_id = header[:16]
    message_id = header[16:19]
    message_type = int.from_bytes(header[19:20], byteorder="big")
    return Header(conversation_id, message_id, message_type)


def serialize_data(data: Any) -> bytes:
    """Turn `data` into a bytes object.

    Due to json serialization, data must not contain a bytes object!
    """
    if isinstance(data, json_objects):
        return data.model_dump_json().encode()  # type: ignore
    else:
        return json.dumps(data, separators=(',', ':')).encode()


def deserialize_data(content: bytes) -> Any:
    """Turn received message content into python objects."""
    return json.loads(content.decode())


def generate_conversation_id() -> bytes:
    """Generate a conversation_id."""
    return uuid7().bytes


def conversation_id_to_datetime(conversation_id: bytes) -> datetime.datetime:
    seconds_since_epoch = int.from_bytes(conversation_id[:6], byteorder="big", signed=False) / 1000
    return datetime.datetime.fromtimestamp(seconds_since_epoch, tz=datetime.timezone.utc)


def _get_json_object_type(data: dict[str, Any]) -> JsonContentTypes:
    if isinstance(data, dict):
        if "method" in data.keys():
            return JsonContentTypes.REQUEST
        elif "result" in data.keys():
            return JsonContentTypes.RESULT_RESPONSE
        elif "error" in data.keys():
            return JsonContentTypes.ERROR_RESPONSE
    return JsonContentTypes.INVALID


def get_json_content_type(data: Any) -> JsonContentTypes:
    if isinstance(data, list):
        content = JsonContentTypes.BATCH if data else JsonContentTypes.INVALID
        for element in data:
            element_typ = _get_json_object_type(element)
            if element_typ == JsonContentTypes.INVALID:
                return JsonContentTypes.INVALID
            else:
                content |= element_typ
        return content
    else:
        return _get_json_object_type(data)
